/** PRQueue - queue with position reservation for multi-threaded programs
**
** Copyright (C) 2009  Eugene Surman
** You may distribute under the terms of the GNU General Public License
**
** Methods and functions :
**
** QueueTest::init
** void QueueTest::start_input_thread
** QueueTest::start_output_thread
** StringMsg::process
** process_msg
** print_msg
** main
**/

#include <prqueue.hpp>

#define MAX_THREADS 8

static void* process_msg( void* arg);
static void* print_msg( void* arg);

static int Thidx =0;

//------------
void QueueTest::init( int qty_threads)
{ 
    if( m_inited) return;

    if( qty_threads > MAX_THREADS || qty_threads <1 ) {
	cerr << "QueueTest: number of threads to start, should be >=1 and <=" << MAX_THREADS << " -exit" << endl;
	exit( 1);
    }
    m_qty_threads =qty_threads;
	
    start_output_thread();
    for( int i=1; i < m_qty_threads +1; i++)
	start_input_thread();

    m_inited =true;
    return;
}//


//------------
void QueueTest::start_input_thread()
{
    int err;
    pthread_t tid;
    if( (err =pthread_create( &tid, NULL, &process_msg, this))) {
        cerr << "QueueTest: could not create input log thread, err=" << strerror( err) << " -exit" << endl;
        exit( 1);
    }
    
    return;
}

//------------
void QueueTest::start_output_thread()
{
    int err;
    pthread_t tid;
    if( (err =pthread_create( &tid, NULL, &print_msg, this))) {
        cerr << "QueueTest: could not create output log thread, err=" << strerror( err) << " -exit" << endl;
        exit( 1);
    }
    return;
}//


//------------
static void* process_msg( void* arg)
{
    int thidx = ++Thidx;
    QueueTest* quetest =(QueueTest*)arg;
    Msg* msg;
    PRQueue< Msg*>::position pos;
    
    cout << "Input thread=" << thidx << " started" << endl;
    for( ;;)
    {
	// Wait for next message appeared in input queue,
	// pop it up and get push position allocated in output queue
	quetest->input_que.pop( msg, quetest->output_que, pos);
	
	// Process message
	msg->process( thidx);
	
	// Push processed message into output queue using acquired position	
	quetest->output_que.push( msg, pos);
    }    
    
    return NULL;
}//


//------------
static void* print_msg( void* arg)
{
    QueueTest* quetest =(QueueTest*)arg;
    Msg* msg;

    cout << "Output thread started" << endl;
    for( ;;)
    {
	quetest->output_que.pop( msg);
	msg->print();
	delete msg;
    }    
    
    return NULL;
}//


#ifdef LONG_MSG  // Try long-time process
//------------
void StringMsg::process( int thidx)
{
    tm tmval;
    StringMsg tmp_msg;
    
    localtime_r( &m_time.tv_sec, &tmval);

    double fld1=123456.789, fld2=223456.789, fld3=323456.789, fld4=423456.789
           ,fld5=523456.789, fld6=623456.789, fld7=723456.789, fld8=823456.789
           ,fld9=923456.789, fld10=1023456.789, fld11=1123456.789, fld12=1123456.789
           ,fld13=1323456.789, fld14=1423456.789, fld15=1523456.789, fld16=1523456.789;
    
    snprintf( tmp_msg.m_buf, tmp_msg.size(), "%d: %02d%02d%2d %02d:%02d:%02d.%06d : %s"
", fld1=%f, fld2=%f, fld3=%f, fld4=%f, fld5=%f, fld6=%f, fld7=%f, fld8=%f"
", fld9=%f, fld10=%f, fld11=%f, fld12=%f, fld13=%f, fld14=%f, fld15=%f, fld16=%f\n"
	,thidx, tmval.tm_year -100, tmval.tm_mon +1, tmval.tm_mday, tmval.tm_hour
	,tmval.tm_min, tmval.tm_sec, (int)m_time.tv_usec, m_buf
	,fld1, fld2, fld3, fld4, fld5, fld6, fld7, fld8
	,fld9, fld10, fld11, fld12, fld13, fld14, fld15, fld16);
	
    strncpy( m_buf, tmp_msg.m_buf, tmp_msg.size());
    
    return;
}//

#else // SHORT_MSG
//------------ 
void StringMsg::process( int thidx)
{
    tm tmval;
    StringMsg tmp_msg;

    localtime_r( &m_time.tv_sec, &tmval);
    snprintf( tmp_msg.m_buf, tmp_msg.size(), "%d: %02d%02d%2d %02d:%02d:%02d.%06d : %s\n"
	,thidx, tmval.tm_year -100, tmval.tm_mon +1, tmval.tm_mday, tmval.tm_hour
	,tmval.tm_min, tmval.tm_sec, (int)m_time.tv_usec, m_buf);
	
    strncpy( m_buf, tmp_msg.m_buf, tmp_msg.size());
    
    return;
}//

#endif


//------------
int main( int argc, char** argv)
{
    int nthreads =3, nmsgs;

    if( argc <2 || (nmsgs =atoi( argv[1])) <=0 ) {
	cout << "usage: " << argv[0] << " nmsgs" << endl;
	return 1;
    }

#ifdef SIMPLE_QUE        
    const char* txt ="SimpleQueue";
#else
    const char* txt ="PRQueue";
#endif
        
    cout << "Start " << txt << " test, num_of_threads=" << nthreads
         << ", num_of_msgs=" << nmsgs << " ..." << endl;
	 
    QueueTest quetest( nthreads);
    
    usleep( 1000);
    cout << "-------------------------------------" << endl;
    int i1 =0;
    for( int i =nmsgs; i; i--) {
	quetest.push( "| %d", i1++ );
	quetest.push( "- %d", i1++ );
    }

    while( ! quetest.empty()) {  
	usleep( 1000);
    }   
    cout << "-------------------------------------" << endl;
    cout << "End " << txt << " test, num_of_threads=" << nthreads
	 << ", num_of_msgs=" << nmsgs << endl;
    
    return 0;
}

// prqueue.cpp


